"""Module configuration objects."""

from django.apps import AppConfig
from django.conf import settings
from django.db import DatabaseError

from . import utils


def _set(key, default):
    key = 'ABOUTCONFIG_%s' % key
    setattr(settings, key, getattr(settings, key, default))


class AboutconfigConfig(AppConfig):
    """Configuration provider for the module.

    Ensures configuration is loaded into cache on application start-up."""

    name = 'aboutconfig'


    @classmethod
    def ready(cls):
        _set('CACHE_NAME', 'default')
        _set('CACHE_TTL', None)
        _set('AUTOLOAD', True)

        # can't load data if models don't exist in the db yet
        if settings.ABOUTCONFIG_AUTOLOAD and cls.migrations_applied():
            utils.preload_cache()


    @classmethod
    def migrations_applied(cls):
        """Check if module's migrations have been applied yet."""

        from django.db.migrations.loader import MigrationLoader
        from django.db import connection

        try:
            loader = MigrationLoader(connection, ignore_no_migrations=True)
            return (cls.name, '0002_initial_data') in loader.applied_migrations
        except DatabaseError: # pragma: no cover
            # errors can happen if database does not exist yet
            return False



"""
Provides the default serializer implementations that come with the module.

See the BaseSerializer documentation for details on how to implement your own.
"""

from decimal import Decimal, InvalidOperation
import re

from django.utils import six
from django.core.exceptions import ValidationError


class BaseSerializer(object):
    """
    Sample class providing the serializer interface.

    You don't have to extend this class to create a valid serializer as long as your
    custom class has the following methods: ``serialize(val)``, ``unserialize(val),``
    ``validate(val)``.

    The constructor should also accept the current configuration object as the first argument.
    This is useful if you want to implement a serializer whose output depends on some
    other configuration data.
    """

    def __init__(self, config):
        """
        Constructor.

        Requires an instance of aboutconfig.models.Config.
        """

        self.config = config

    @staticmethod
    def is_class_valid(klass):
        """
        Verify that the given class object has all the required methods. This check is performed
        when the class is used to process the data.

        Checks for existence of ``serialize()``, ``unserialize()``, ``validate()``.
        """

        try:
            assert hasattr(klass, 'serialize')
            assert hasattr(klass, 'unserialize')
            assert hasattr(klass, 'validate')
        except AssertionError:
            return False
        else:
            return True

    def serialize(self, val):
        """
        Convert a python value to string.

        Dummy implementation. Override in subclasses.
        """

        raise NotImplementedError()

    def unserialize(self, val):
        """
        Convert a string value to a python object.

        Dummy implementation. Override in subclasses.
        """

        raise NotImplementedError()

    def validate(self, val):
        """
        Validate a serialized value. Primarily used in forms (which return strings).

        This method must raise a ``django.core.exceptions.ValidationError`` if the value
        does not pass validation.

        Dummy implementation. Optionally override in subclasses.
        """

        pass


class StrSerializer(BaseSerializer):
    """Built-in serializer for strings."""

    def serialize(self, val):
        """Does essentially nothing since the serialized value is the original value."""

        return six.text_type(val)

    def unserialize(self, val):
        """Does essentially nothing since the serialized value is the original value."""

        return six.text_type(val)


class IntSerializer(BaseSerializer):
    """Built-in serializer for integers."""

    def serialize(self, val):
        """Convert an integer to string."""

        return six.text_type(val)

    def unserialize(self, val):
        """Convert a string representation into an integer."""

        return int(val)

    def validate(self, val):
        """
        Assert string is a valid integer representation.

        Raises ``ValidationError``.
        """

        if not re.match(r'^-?\d+$', val):
            raise ValidationError('Not a valid integer')


class BoolSerializer(BaseSerializer):
    """Built-in serializer for boolean values."""

    def serialize(self, val):
        """Convert a boolean into a string."""

        return 'true' if val else 'false'

    def unserialize(self, val):
        """Convert a string representation into a boolean."""

        val = val.lower()
        return val == 'true'

    def validate(self, val):
        """
        Assert string is a valid boolean representation.

        Raises ``ValidationError``.
        """

        if val.lower() not in ('true', 'false'):
            raise ValidationError('Must be "true" or "false"')


class DecimalSerializer(BaseSerializer):
    """Built-in serializer for decimal.Decimal objects."""

    def serialize(self, val):
        """Convert a decimal object into a string."""

        return six.text_type(val)

    def unserialize(self, val):
        """Convert a string representation into a decimal object."""

        return Decimal(val)

    def validate(self, val):
        """
        Assert string is a valid decimal representation.

        Raises ``ValidationError``.
        """

        try:
            Decimal(val)
        except InvalidOperation:
            raise ValidationError('Not a valid decimal')



"""Common utilities used by the module."""

import importlib
from collections import namedtuple

from django.core.exceptions import ValidationError
from django.core.cache import caches
from django.conf import settings

from .serializers import BaseSerializer
from .constants import CACHE_KEY_PREFIX

_SENTINEL = object()
DataTuple = namedtuple('DataTuple', 'value,allow_template_use')


def _cache_key_transform(key):
    """Transforms the configuration key to avoid undesired collisions in cache."""

    return CACHE_KEY_PREFIX + key.lower()


def _get_cache():
    """Return the configured cache backend."""

    return caches[settings.ABOUTCONFIG_CACHE_NAME]


def _set_cache(config):
    """Put the given configuration data into cache."""

    cache = _get_cache()
    cache_key = _cache_key_transform(config.key)
    cache.set(cache_key, DataTuple(config.get_value(), config.allow_template_use),
              settings.ABOUTCONFIG_CACHE_TTL)


def _delete_cache(config):
    """Remove value from cache."""

    cache = _get_cache()
    cache_key = _cache_key_transform(config.key)
    cache.delete(cache_key)


def load_class(class_path):
    """Load class from absolute class path."""

    split_path = class_path.split('.')
    class_name = split_path.pop()
    module_path = '.'.join(split_path)

    module = importlib.import_module(module_path)
    return getattr(module, class_name)


def load_serializer(class_path):
    """Load a class by name/path that implements the serialization interface.

    Returns an instance of the class or raises ImportError/ValueError/AttributeError."""

    klass = load_class(class_path)

    if not BaseSerializer.is_class_valid(klass):
        raise ValueError('"{}" is not a valid serializer'.format(class_path))

    return klass


def serializer_validator(class_path):
    """Check whether the given class (via path) is a valid serialzier or not.

    Raises ValidationError if not valid."""

    try:
        load_serializer(class_path)
    except (ValueError, ImportError, AttributeError):
        raise ValidationError('Invalid serializer class')


def get_config(key, value_only=True):
    """Get the configuration value for the given key.

    By default this function only returns the value, set value_only=False to get a
    2-tuple with the following elements: (value, allow_template_use). This functionality
    is almost certainly not needed by end-users, only included for the built-in template
    tag filtering functionality.

    Returns None if no such configuration exists (or (None, True) if value_only=False)."""

    from .models import Config

    cache = _get_cache()
    cache_key = _cache_key_transform(key)
    data = cache.get(cache_key, _SENTINEL)

    if data is _SENTINEL:
        try:
            config = Config.objects.select_related('data_type').get(key=key.lower())
        except Config.DoesNotExist:
            data = DataTuple(None, True)
            cache.set(cache_key, data, settings.ABOUTCONFIG_CACHE_TTL)
        else:
            data = DataTuple(config.get_value(), config.allow_template_use)
            _set_cache(config)

    return data.value if value_only else data


def preload_cache():
    """Load all configuration data into cache."""

    from .models import Config

    for config in Config.objects.all():
        _set_cache(config)



"""User configuration access via the django admin."""

from django.contrib import admin
from django.conf.urls import url
from django.core.exceptions import ValidationError
from django.http import Http404, JsonResponse, HttpResponseBadRequest

from .models import DataType, Config
from .forms import ConfigAdminForm
from .constants import CONFIG_ADMIN_TYPE_QUERY_KEY


@admin.register(DataType)
class DataTypeAdmin(admin.ModelAdmin):
    """Django admin for the DataType model."""

    list_display = ('name', 'serializer_class')


class ConfigNamespaceFilter(admin.SimpleListFilter):
    """Key namespace sidebar filter implementation."""

    title = 'namespace'
    parameter_name = 'namespace'


    def lookups(self, request, model_admin):
        queryset = model_admin.get_queryset(request)
        lookups = set(queryset.values_list(
            'key_namespace', flat=True
        ).distinct().iterator())

        return sorted((x, x) for x in lookups)


    def queryset(self, request, queryset):
        if self.value() is None:
            return queryset

        return queryset.filter(key_namespace=self.value())


@admin.register(Config)
class ConfigAdmin(admin.ModelAdmin):
    """Django admin for the Config model."""

    list_display = ('key', 'value', 'default_value', 'data_type', 'allow_template_use', 'in_cache')
    list_filter = ('data_type', 'allow_template_use', ConfigNamespaceFilter)
    search_fields = ('key',)
    form = ConfigAdminForm

    class Media:
        js = ('aboutconfig/config_admin.js',)


    @staticmethod
    def fetch_field_view(request):
        """Fetches the HTML required for rendering the given type of value."""

        try:
            data_type = DataType.objects.get(pk=request.GET[CONFIG_ADMIN_TYPE_QUERY_KEY])
        except (DataType.DoesNotExist, ValueError, TypeError, KeyError):
            raise Http404()

        try:
            attrs = {'id': request.GET['id']}
        except KeyError:
            return HttpResponseBadRequest()

        current_value = request.GET.get('value', '')

        try:
            config = Config.objects.get(pk=request.GET['config_pk'])
        except (DataType.DoesNotExist, ValueError, TypeError, KeyError):
            config = Config(key='x', key_namespace='x')

        config.data_type = data_type
        config.value = current_value

        try:
            config.full_clean()
        except ValidationError:
            current_value = ''

        widget = data_type.get_widget_class()(**data_type.widget_args)
        return JsonResponse({'content': widget.render('value', current_value, attrs=attrs)})


    def get_urls(self):
        """Get custom admin endpoints."""

        return [
            url(r'^fetch-value-field-html',
                self.admin_site.admin_view(self.fetch_field_view),
                name='ac-fetch-value-field')
        ] + super(ConfigAdmin, self).get_urls()



"""Provides ability to (re)load config cache from the command line."""

from django.core.management.base import BaseCommand

from aboutconfig import utils

class Command(BaseCommand):
    """Loads all existing config entries into cache"""

    help = 'Loads all existing config entries into cache'

    def handle(self, *args, **options):
        utils.preload_cache()



"""Forms for model editing."""

from django import forms
from django.core.urlresolvers import reverse

from .models import Config, DataType


class ConfigAdminForm(forms.ModelForm):
    """Form used for editing Config objects in the django admin."""

    def __init__(self, *args, **kwargs):
        """Constructor.

        Replaces the default widgets with new ones."""

        super(ConfigAdminForm, self).__init__(*args, **kwargs)

        self.fields['data_type'].widget = forms.Select(
            choices=self.fields['data_type'].widget.choices,
            attrs={
                'class': 'aboutconfig-type-field',
                'data-url': reverse('admin:ac-fetch-value-field'),
                'data-instance-pk': self.instance.pk or ''
            }
        )

        data_type = self.get_data_type()

        original_field = self.fields['value']
        field_class = original_field.__class__
        self.fields['value'] = field_class(
            initial=original_field.initial,
            widget=data_type.get_widget_class()(**data_type.widget_args),
        )


    class Meta:
        model = Config
        fields = ('data_type', 'key', 'value', 'allow_template_use')
        readonly_fields = ('default_value',)


    def get_data_type(self):
        """Get data type to use when figuring out how to render the value field."""

        try:
            return self.instance.data_type
        except DataType.DoesNotExist:
            return DataType.objects.get(name='String')



from __future__ import unicode_literals

from django.db import migrations

from ..serializers import BoolSerializer, IntSerializer, StrSerializer, DecimalSerializer

def get_path(klass):
    return '{}.{}'.format(klass.__module__, klass.__name__)


def load_data_types(apps, schema_editor):
    DataType = apps.get_model('aboutconfig', 'DataType')

    DataType.objects.create(name='Boolean', serializer_class=get_path(BoolSerializer),
                            widget_class='aboutconfig.widgets.Select',
                            widget_args_raw='{"choices": [["true", "True"], ["false", "False"]]}')
    DataType.objects.create(name='Integer', serializer_class=get_path(IntSerializer),
                            widget_class='aboutconfig.widgets.NumberInput')
    DataType.objects.create(name='String', serializer_class=get_path(StrSerializer))
    DataType.objects.create(name='Decimal', serializer_class=get_path(DecimalSerializer))


class Migration(migrations.Migration):

    dependencies = [
        ('aboutconfig', '0001_initial'),
    ]

    operations = [
        migrations.RunPython(load_data_types)
    ]



"""Aliases for the default django widgets.

In case of Django ever moving them around, only this file will need updating. No new migrations."""

# pylint: disable=unused-import
from django.forms import Select, NumberInput



"""A firefox-like about:config implementation for one-off settings in Django apps."""

__version__ = (0, 5, 3)
default_app_config = 'aboutconfig.apps.AboutconfigConfig' # pylint: disable=invalid-name


def get_config(key, value_only=True):
    """Get configured value by key.

    By default returns value only. If ``value_only`` is ``False``, returns an instance of
    aboutconfig.utils.DataTuple which also contains the ``allow_template_use`` value.

    This is a lazy wrapper around the internal ``utils.get_config()`` function."""

    from . import utils

    return utils.get_config(key, value_only)



"""Database models used by this module."""

import json

from django.forms import TextInput
from django.db import models
from django.dispatch import receiver
from django.core.exceptions import ValidationError
from django.core.validators import RegexValidator
from django.utils.encoding import python_2_unicode_compatible
from django.utils.translation import ugettext_lazy as _

from .constants import KEY_REGEX
from . import utils


__all__ = ('DataType', 'Config')


@python_2_unicode_compatible
class DataType(models.Model):
    """
    Model representing the data type.

    Each model instance is associated with a serializer class that performs the casting
    of python objects to strings and back.
    """

    name = models.CharField(max_length=32, verbose_name=_('Name'))
    serializer_class = models.CharField(
        max_length=256, validators=[utils.serializer_validator], verbose_name=_('Serializer class'),
        help_text=_('Must be a class that implements serialize, unserialize and validate methods.'))
    widget_class = models.CharField(
        max_length=256, blank=True, verbose_name=_('Widget class'),
        help_text=_('Widget class used to edit values of this data type. TextInput by default.'))
    widget_args_raw = models.CharField(
        max_length=1024, default='{}', verbose_name=_('Raw widget arguments'),
        help_text=_('Additional data for the value field widget (JSON).'))


    class Meta:
        verbose_name = _('Data-type')


    @property
    def widget_args(self):
        """Parse widget args value."""
        return json.loads(self.widget_args_raw)


    @widget_args.setter
    def widget_args(self, val):
        """Serialize widget args value."""
        self.widget_args_raw = json.dumps(val)


    def get_class(self):
        """Load and the configured serializer class."""

        return utils.load_serializer(self.serializer_class)


    def get_widget_class(self):
        """Load the widget configured class."""

        if not self.widget_class:
            return TextInput
        else:
            return utils.load_class(self.widget_class)


    def __str__(self):
        return self.name


@python_2_unicode_compatible
class Config(models.Model):
    """
    Model representing the user-defined configuration data.

    Most of the time just key, value and data_type need to be set. The default_value field is
    used for by 3rd-party applications that rely on this module that require that some sane value
    is set in the configuration. Users can override this if wish, but the default value remains.

    The allow_template_use field is only used in the get_config template filter. There are times
    when you have some sensitive configuration stored, but want to prevent it from potentially
    leaking out via user-configurable templates. If the value is set to false, the filter will act
    as if the configuration does not exist at all.

    Note: an empty string is equivalent to None, and the configuration will fall back to the
    default value.
    """

    EMPTY_VALUES = (None, '')

    key = models.CharField(
        max_length=512, validators=[RegexValidator(KEY_REGEX)], unique=True,
        verbose_name=_('Key'),
        help_text=_('Period separated strings. All keys are case-insensitive.'))
    key_namespace = models.CharField(
        max_length=512, db_index=True, verbose_name=_('Key namespace'))
    value = models.CharField(
        max_length=1024, blank=True, null=True, verbose_name=_('Value'))
    data_type = models.ForeignKey(DataType, related_name='+', verbose_name=_('Data-type'))
    default_value = models.CharField(
        max_length=1024, editable=False, verbose_name=_('Default value'),
        help_text=_('Default value set by setting provider. Used by 3rd-party apps.'))
    allow_template_use = models.BooleanField(
        default=True, verbose_name=_('Allow template use'),
        help_text=_('Prevent settings from being accessible via the template filter. Can ' \
                    'be useful for API-keys, for example'))


    class Meta:
        ordering = ('key', 'value', 'default_value')
        verbose_name = _('Config')


    def save(self, **kwargs):
        """Save model to database."""

        self.key = self.key.lower()
        self.key_namespace = self.key.split('.')[0]
        super(Config, self).save(**kwargs) # pylint: disable=no-member


    def full_clean(self, **kwargs):
        """Validate model data before saving."""

        super(Config, self).full_clean(**kwargs) # pylint: disable=no-member

        if self.value in self.EMPTY_VALUES:
            self.value = None

            if self.default_value is None:
                raise ValidationError({'value': ValidationError(_('A value is required'))})
            else:
                # have default value to fall back on
                return

        try:
            self._get_serializer().validate(self.value)
        except ValidationError as e:
            raise ValidationError({'value': e})


    def get_raw_value(self):
        """
        Get serialized value.

        Tries to get manually set value before falling back to default value.
        """

        if self.value in self.EMPTY_VALUES:
            return self.default_value
        else:
            return self.value


    def get_value(self):
        """Get unserialized value."""

        return self._get_serializer().unserialize(self.get_raw_value())


    def set_value(self, val):
        """Set the configuration value via a python object. Takes care of the serialization."""

        self.value = self._get_serializer().serialize(val)


    def _get_serializer(self):
        """Get the configured serializer class instance."""

        return self.data_type.get_class()(self)


    # pylint: disable=protected-access
    def in_cache(self):
        """
        Check if the configuration value is already present in cache.

        Works like a python get_config() call - ignores allow_template_use setting.
        """

        key = utils._cache_key_transform(self.key)
        return utils._get_cache().has_key(key)
    in_cache.boolean = True # django admin icon fix


    def __str__(self):
        return '{}={}'.format(self.key, self.get_raw_value())


# pylint: disable=unused-argument
@receiver(models.signals.post_save, sender=Config)
def update_cache(instance, **kwargs):
    """
    Update cache for the given configuration instance.

    Automatically called on configuration creation/update.
    """

    # pylint: disable=protected-access
    utils._set_cache(instance)


# pylint: disable=unused-argument
@receiver(models.signals.post_delete, sender=Config)
def delete_cache(instance, **kwargs):
    """
    Remove cache for the given configuration instance.

    Automatically called on configuration deletion.
    """

    # pylint: disable=protected-access
    utils._delete_cache(instance)



"""
Template tags and filters provided by the module.

Add "{% load config %}" in your templates to access.
"""

from django import template
from django.template.defaultfilters import stringfilter

from .. import utils

# pylint: disable=invalid-name
register = template.Library()


def _get_config_for_template(key):
    """Get the configuration data while respecting the template use flag."""

    data = utils.get_config(key, value_only=False)

    if data.allow_template_use and data.value is not None:
        return data.value

    return '' # returning None makes django print "None" which is undesirable


@register.filter(name='get_config')
@stringfilter
def get_config_filter(key):
    """
    Get the configuration value for the given key.

    If allow_template_use is False, will act as if the key is not set.

    Use in combination with the built-in `default` filter for default values.
    """

    return _get_config_for_template(key)


@register.assignment_tag(name='get_config')
def get_config_assignment_tag(key):
    """
    Get the configuration value for the given key.

    Assigns the value to another variable.
    """

    return _get_config_for_template(key)



# -*- coding: utf-8 -*-

import os
from setuptools import setup, find_packages

from aboutconfig import __version__

BASE_DIR = os.path.dirname(os.path.abspath(__file__))
VERSION_STRING = '.'.join(str(s) for s in __version__)

try:
    import pypandoc
    long_description = pypandoc.convert('README.md', 'rst')
except(IOError, ImportError):
    with open(os.path.join(BASE_DIR, 'README.md')) as fp:
        long_description = fp.read()

setup(
    name='django-aboutconfig',
    version=VERSION_STRING,
    url='https://bitbucket.org/impala/django-aboutconfig',
    license='GPLv3+',
    description='A firefox-like about:config implementation for one-off settings in Django apps.',
    long_description=long_description,
    keywords=['django', 'library', 'configuration'],
    author='Kirill Stepanov',
    author_email='mail@kirillstepanov.me',
    packages=find_packages(),
    download_url='https://bitbucket.org/impala/django-aboutconfig/get/%s.tar.gz' % VERSION_STRING,
    classifiers=[
        'Development Status :: 4 - Beta',
        'Framework :: Django',
        'Framework :: Django :: 1.8',
        'Framework :: Django :: 1.9',
        'Framework :: Django :: 1.10',
        'Framework :: Django :: 1.11',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Topic :: Internet :: WWW/HTTP',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
    ],
    install_requires=[
        'Django>=1.8',
    ],
    include_package_data=True,
    zip_safe=False,
)
